from manim import *
from manim.constants import DOWN, LEFT, ORIGIN, RIGHT, UP
from manim.mobject.geometry.arc import Circle
from manim.mobject.geometry.polygram import Square, Triangle
from manim.mobject.text.tex_mobject import MathTex, Tex
from manim.mobject.types.vectorized_mobject import VGroup
from manim.utils.tex_templates import TexFontTemplates

# config["pixel_height"] = 500  #  1080 is default
# config["pixel_width"] = 1500  # 1920 is default


class PowerManim(VGroup):
    def build_shapes(self):
        logo_green = "#87c2a5"
        logo_blue = "#525893"
        logo_red = "#e07a5f"

        circle = Circle(color=logo_green, fill_opacity=1).shift(LEFT)

        square = Square(color=logo_blue, fill_opacity=1).shift(UP)
        triangle = Triangle(color=logo_red, fill_opacity=1).shift(RIGHT)

        square = Difference(square, circle, color=logo_blue, fill_opacity=1)
        triangle = Difference(triangle, square, color=logo_red, fill_opacity=1)

        circle.shift(DL * self.gap)
        triangle.shift(DR * self.gap)

        return VGroup(circle, square, triangle)

    def build_text(self):
        ds_m = MathTex(r"\mathbb{PM}", fill_color=self.logo_black).scale(5)
        ds_m.shift(4.25 * LEFT + 1.5 * UP)
        return ds_m

    def build_banner(self):
        shapes = self.build_shapes()

        ds_p = MathTex(r"\mathbb{P}", fill_color=self.logo_black).scale(2)
        ower = Tex("ower", tex_template=TexFontTemplates.gnu_freeserif_freesans).scale(2)
        ower.next_to(ds_p, RIGHT, buff=SMALL_BUFF).align_to(ds_p, DOWN).set_color(self.font_color)
        power = VGroup(ds_p, ower)

        ds_m = MathTex(r"\mathbb{M}", fill_color=self.logo_black).scale(2)
        anim = Tex("anim", tex_template=TexFontTemplates.gnu_freeserif_freesans).scale(2)
        anim.next_to(ds_m, RIGHT, buff=SMALL_BUFF).align_to(ds_m, DOWN).set_color(self.font_color)
        tmanim = VGroup(ds_m, anim)

        banner = VGroup(power, tmanim).arrange(RIGHT, buff=MED_SMALL_BUFF).move_to(ORIGIN).scale(1.5)
        banner.next_to(shapes[1], LEFT)
        banner.align_to(shapes[0].get_critical_point(UP) + self.gap, DOWN)

        return VGroup(shapes, banner).move_to(ORIGIN)

    def __init__(self, font_color="#343434", logo_black="#343434", gap=0.4):
        self.font_color = font_color
        self.logo_black = logo_black
        self.gap = gap

        shapes = self.build_shapes()
        ds_m = self.build_text()
        super().__init__(*shapes, ds_m)
        self.move_to(ORIGIN)



from powermanim.showcase.showcasescene import ShowcaseScene
from powermanim.templates.sectiontitle import SectionTitle


class SectionTitleShowcase(ShowcaseScene):
    def showcasing():
        return SectionTitle

    def construct(self):
        title = SectionTitle("Section Title")
        self.play(title.show())
        self.wait(1)
        self.play(title.hide())
        self.wait(1)



from manim import *

from powermanim.components.vgrouphighlight import VGroupHighlight
from powermanim.showcase.showcasescene import ShowcaseScene


class VGroupHighlightShowcase(ShowcaseScene):
    def showcasing():
        return VGroupHighlight

    def construct(self):
        dots = [
            Dot(radius=0.25, color=color, fill_opacity=0.25).shift(i * RIGHT)
            for i, color in zip(
                range(-2, 3),
                [
                    RED,
                    GREEN,
                    BLUE,
                    YELLOW,
                    PINK,
                ],
            )
        ]

        group = VGroupHighlight(*dots, anim_run_time=1, anim_lag_ratio=0.1)
        self.add(group)
        self.play(group.highlight(0))
        self.play(group.highlight(1))
        self.play(group.highlight([2, 3]))
        self.play(group.highlight([1, 3]))
        self.play(group.highlight([]))
        self.play(group.highlight([2, 4]))
        self.play(group.highlight([]))
        self.wait(0.5)



from manim import *

from powermanim.components.invariance import Invariance
from powermanim.showcase.showcasescene import ShowcaseScene


class InvarianceShowcase(ShowcaseScene):
    def showcasing():
        return Invariance

    def construct(self):
        invshowcase: Invariance = Invariance(
            Axes(
                x_range=[-1.5, 1.5],
                y_range=[-1.5, 1.5],
                x_length=4,
                y_length=4,
                axis_config={"include_ticks": False},
            ).scale(1.5)
        )

        self.add(invshowcase)

        for i in [
            np.pi * 0.1,
            np.pi * 0.15,
            np.pi * 0.2,
        ]:
            self.play(
                invshowcase.transformation_anim(
                    np.array(
                        [
                            [np.cos(i), -np.sin(i), 0],
                            [np.sin(i), np.cos(i), 0],
                            [0, 0, 1],
                        ]
                    ),
                    run_time=0.5,
                )
            )

        for i in range(2, 5):
            self.play(
                invshowcase.transformation_anim(
                    invshowcase.get_random_ortho_matrix(),
                    run_time=0.5,
                )
            )

        point_indices = [0, 1, 2, 7, 8]
        radii_anim, lines = invshowcase.radii_anim(point_indices, run_time=0.275)
        self.play(radii_anim)

        self.play(invshowcase.rescale_along_radius_anim(point_indices, lines=lines))

        self.play(
            invshowcase.rescale_all_to_unit_anim(
                run_time=1.0,
            )
        )



from manim import *

from powermanim.showcase.showcasescene import QUALITY, WOWCHEMY_FOLDER, ShowcaseMeta


def test_and_gen_showcases():
    for scene in ShowcaseMeta.showcase_scenes:
        with tempconfig(
            {
                "quality": QUALITY,
                "disable_caching": True,
                "custom_folders": True,
                "media_dir": WOWCHEMY_FOLDER,
            }
        ):
            scene = scene()
            scene.render()



from manim import *

from powermanim.layouts.arrangedbullets import ArrangedBullets, Bullet
from powermanim.showcase.showcasescene import ShowcaseScene


class ArrangedBulletsShowcase(ShowcaseScene):
    def showcasing():
        return ArrangedBullets

    def construct(self):
        rows = [
            Bullet("First row"),
            Bullet("Second row"),
            Bullet("Elements:"),
            Bullet("First element", level=1, symbol="(1)"),
            Bullet("Second element", level=1, symbol="(2)"),
            Bullet("Third element", level=1, symbol="(3)"),
        ]
        g_rows = VGroup(*rows).set_opacity(0.25).scale(0.9)
        g_rows.target = ArrangedBullets(
            *g_rows.copy(),
            line_spacing=MED_LARGE_BUFF * 1.25,
            left_buff=MED_LARGE_BUFF * 3,
        ).set_opacity(1)

        self.play(
            MoveToTarget(g_rows),
            rate_func=there_and_back_with_pause,
            run_time=3,
        )

        self.wait()



from manim import *


class Reference(VGroup):
    def __init__(self, text: str, font_size: int = 28, color: ManimColor = LIGHT_GREY):
        text = rf"\textit{{{text}}}"
        boxtext = rf"\mbox{{{text}}}"
        tex = Tex(
            boxtext,
            font_size=font_size,
            color=color,
        ).to_corner(DL, buff=MED_SMALL_BUFF)
        super().__init__(tex)



from manim import *
from scipy.special import softmax

from powermanim.components.chartbars import ChartBars
from powermanim.showcase.showcasescene import ShowcaseScene


class ChartBarsShowcase(ShowcaseScene):
    def showcasing():
        return ChartBars

    def construct(self):
        axes = Axes(x_range=[0, 6], y_range=[0, 1.5], x_length=7, y_length=4)
        size = 5
        changes = 3

        dist1 = softmax(np.random.randn(size))

        bars = ChartBars(axes, dist1, xs=list(range(size)), fill_color=RED, stroke_width=0.1)
        self.add(axes, bars)
        for i in range(changes):
            dist2 = softmax(np.random.randn(size))
            self.play(bars.animate.set_values(dist2), run_time=2)
        self.play(bars.animate.set_values(dist1), run_time=2)



import typing as T

from manim import *


class VGroupHighlight(VGroup):
    def __init__(
        self,
        *args: VMobject,
        anim_run_time: float = 1.0,
        anim_lag_ratio: float = 0,
        active_opacity: float = 1.0,
        scale_active: float = 1.0,
        scale_about_point=None,
        scale_about_edge=None,
        **kwargs: VMobject,
    ) -> None:
        """Group component that can highlight a subset of its submobjects.

        Args:
            *args: Submobjects to be added to the VGroup.
            anim_run_time (float): The run time of the animation.
            anim_lag_ratio (float): The lag ratio of the animation.
            active_opacity (float): The opacity of the highlighted submobjects.
            scale_active (float): The scale of the highlighted submobjects.
            scale_about_point (np.ndarray): The point to scale about.
            scale_about_edge (np.ndarray): The edge to scale about.
            **kwargs: Keyword arguments to be passed to the VGroup.
        """
        super().__init__(*args, **kwargs)

        self.anim_run_time = anim_run_time
        self.anim_lag_ratio = anim_lag_ratio
        self.active_opacity = active_opacity
        self.scale_active = scale_active
        self.scale_about_point = scale_about_point
        self.scale_about_edge = scale_about_edge

        self.previously_active_idxs = []

    def highlight(self, indices: T.Union[int, T.Sequence[int]]) -> AnimationGroup:
        """Highlights the submobjects in the given indices in the scene.

        Args:
            scene:
                The scene in which the animation is played.

            indices:
                The indices to highlight. If a single integer is given, only that index is highlighted.
                If a sequence of integers is given, all indices in the sequence are highlighted. The
                previously highlighted indices are dehighlighted smoothly.
        """
        anims = []

        if not isinstance(indices, T.Sequence):
            indices = [indices]

        for to_highlight in indices:
            item = self.submobjects[to_highlight]
            if not hasattr(item, "saved_state") or item.saved_state is None:
                item.save_state()
            item.target = item.saved_state.copy()
            item.target.scale(
                self.scale_active,
                about_point=self.scale_about_point,
                about_edge=self.scale_about_edge,
            )
            item.target.set_opacity(self.active_opacity)
            anims.append(MoveToTarget(item))

        if self.previously_active_idxs:
            for previously_active_idx in self.previously_active_idxs:
                if previously_active_idx not in indices:
                    anims.append(
                        self.submobjects[previously_active_idx].animate.restore(),
                    )

        self.previously_active_idxs = indices

        return AnimationGroup(*anims, lag_ratio=self.anim_lag_ratio)



from typing import Tuple

from manim import *
from scipy.stats import ortho_group


class Invariance(VGroup):
    COLORS = [
        ORANGE,
        BLUE,
        TEAL,
        GREEN,
        YELLOW,
        GOLD,
        RED,
        MAROON,
        PURPLE,
        PINK,
    ]

    def __init__(self, axis):
        np.random.seed(seed=1)
        self.active_opacity = 0.8
        self.inactive_opacity = 0.25

        self.axis = axis
        self.points_coords_orig = np.random.rand(17, 3) * 2 - 1
        self.points_coords_orig[:, -1] = 0
        self.points_coords = self.axis.coords_to_point(self.points_coords_orig)

        self.dots = [
            Dot(
                self.points_coords[i],
                fill_opacity=self.active_opacity,
                color=self.COLORS[i % len(self.COLORS)],
                z_index=1,
            )
            for i in range(len(self.points_coords))
        ]

        super().__init__(self.axis, *self.dots)

    def transformation_anim(self, transformation: np.ndarray, bias: np.ndarray = 0, run_time: float = 1.0):
        self.points_coords = self.axis.coords_to_point(np.matmul(self.points_coords_orig, transformation) + bias)

        return AnimationGroup(
            *(self.dots[i].animate.move_to(self.points_coords[i]) for i in range(len(self.points_coords))),
            run_time=run_time,
        )

    def get_random_ortho_matrix(self):
        random_ortho = np.zeros((3, 3))
        random_ortho[:2, :2] = ortho_group.rvs(2)
        return random_ortho

    def radii_anim(self, point_indices, line_overshoot: float = 1.25, run_time=1.0) -> Tuple[Animation, List[Line]]:
        lines = []
        for point_index in point_indices:
            point = self.axis.p2c(self.points_coords[point_index])
            point = point / np.linalg.norm(point)

            line = DashedLine(self.axis.coords_to_point(*ORIGIN), self.axis.c2p(*(point * line_overshoot))).set_opacity(
                0.5
            )
            lines.append(line)

        return AnimationGroup(*[Create(line) for line in lines], run_time=run_time), lines

    def rescale_along_radius_anim(
        self, point_indices, n_iters=3, max_radius=0.95, min_radius=0.2, lines=[]
    ) -> AnimationGroup:
        orig_points = [self.axis.p2c(self.points_coords[point_index]) for point_index in point_indices]

        hide_others = AnimationGroup(
            *[
                self.dots[point_index].animate.set_opacity(self.inactive_opacity)
                for point_index in range(len(self.dots))
                if point_index not in point_indices
            ],
        )

        move_some = []
        for point_idx, point in zip(point_indices, orig_points):
            point_traj = np.tile(point[None], (2 * n_iters + 2, 1))
            point_traj[1:-1, :] /= np.linalg.norm(point_traj, axis=-1, keepdims=True)[1:-1]
            point_traj[1:-1, :] *= np.tile(np.asarray([max_radius, min_radius]), n_iters)[..., None]

            path = VMobject()
            path.set_points_smoothly(self.axis.coords_to_point(point_traj))
            move_some.append(MoveAlongPath(self.dots[point_idx], path, run_time=1.75))

        show_others = AnimationGroup(VGroup(*self.dots).animate.set_opacity(self.active_opacity))

        animations = [
            AnimationGroup(hide_others, run_time=1),
            AnimationGroup(*move_some, lag_ratio=0.05, run_time=n_iters),
        ]

        if lines:
            animations.append(
                AnimationGroup(
                    *(Uncreate(x) for x in lines),
                    show_others,
                    lag_ratio=0.5,
                    run_time=1,
                ),
            )
        else:
            animations.append(AnimationGroup(show_others, run_time=1))
        return Succession(*animations)

    def rescale_all_to_unit_anim(self, run_time=2.0) -> AnimationGroup:
        return AnimationGroup(
            *(
                self.dots[i].animate.move_to(
                    self.axis.c2p(
                        *(self.axis.p2c(self.points_coords[i]) / np.linalg.norm(self.axis.p2c(self.points_coords[i])))
                    )
                )
                for i in range(len(self.points_coords))
            ),
            run_time=run_time,
        )



from manim import ManimColor as Color
from manim import *


class SectionTitle(VGroup):
    def __init__(
        self,
        section_title: str,
        in_run_time: float = 0.75,
        out_run_time: float = 0.25,
        out_shift: float = UP,
        border_color: Color = BLUE,
        border_buff: float = MED_LARGE_BUFF,
        border_lag_ratio: float = 0.5,
    ):
        """Section title transition.

        Args:
            section_title (str): Title of the section
            in_run_time (float, optional): Run time of the animation. Defaults to 0.75.
            out_run_time (float, optional): Run time of the animation. Defaults to 0.25.
            out_shift (float, optional): Shift of the animation. Defaults to UP.
            border_color (Color, optional): Color of the border. Defaults to BLUE.
            border_buff (float, optional): Buffer of the border. Defaults to MED_LARGE_BUFF.
            border_lag_ratio (float, optional): Lag ratio of the border. Defaults to 0.5.
        """
        self.slide_title = Tex(section_title, font_size=72)
        self.border_color = border_color
        self.border_buff = border_buff
        self.border_lag_ratio = border_lag_ratio
        self.in_run_time = in_run_time
        self.out_run_time = out_run_time
        self.out_shift = out_shift

        super().__init__(self.slide_title)

    def show(self) -> Animation:
        """Show the section title.

        Args:
            scene (Scene): the scene.
        """
        return AnimationGroup(
            DrawBorderThenFill(self.slide_title, run_time=self.in_run_time),
            Circumscribe(
                self.slide_title,
                color=self.border_color,
                buff=self.border_buff,
            ),
            lag_ratio=self.border_lag_ratio,
        )

    def hide(self) -> Animation:
        """Hide the section title.

        Args:
            scene (Scene): the scene.
        """
        return AnimationGroup(
            FadeOut(self.slide_title, shift=self.out_shift),
            run_time=self.out_run_time,
        )



try:
    from ._version import __version__ as __version__
except ImportError:
    import sys

    print(
        "Project not installed in the current env, activate the correct env or install it with:\n\tpip install -e .",
        file=sys.stderr,
    )
    __version__ = "unknown"



from manim import *

from powermanim.components.numberslider import NumberSlider
from powermanim.showcase.showcasescene import ShowcaseScene


class NumberSliderShowcase(ShowcaseScene):
    def showcasing():
        return NumberSlider

    def construct(self):
        slider = NumberSlider()
        self.add(slider)
        self.play(slider.tracker.animate.set_value(2), run_time=1)
        self.play(slider.tracker.animate.set_value(-1), run_time=1)
        self.play(slider.animate.shift(LEFT))
        self.play(slider.tracker.animate.set_value(0.5), run_time=1)
        self.play(slider.tracker.animate.set_value(-1.5), run_time=1)
        self.play(slider.tracker.animate.set_value(0), run_time=1)



import itertools
from typing import Iterable, Sequence

from manim import ManimColor as Color
from manim import *


class ChartBars(VGroup):
    def __init__(
        self,
        axes,
        values: Sequence[float | int],
        xs: Sequence[float] | None = None,
        width_ratio: float = 1.0,
        offset: float = 0.5,
        fill_color: Color = BLUE,
        fill_opacity: float = 0.5,
        stroke_color: Color = WHITE,
        stroke_width: float = 1.0,
    ):
        """Create a bar chart from values.

        Args:
            axes: The axes to plot the bars on.
            values: The values to plot. There should be one value for each bar.
            xs: The x-coordinates of the bars. If None, the bars will be centered on
                each integer from 1 to the number of values.
            width_ratio: The width of each bar as a fraction of the distance between x-coordinates.
            offset: The distance of the bars from the x-coordinates as a fraction of the distance between x-coordinates.
            fill_color: The fill color of the bars.
            fill_opacity: The fill opacity of the bars.
            stroke_color: The stroke color of the bars.
            stroke_width: The stroke width of the bars.
        """
        xs = xs if xs is not None else np.arange(*axes.x_range)

        self.x_to_index = dict(zip(xs, itertools.count()))
        x_step = xs[1] - xs[0]
        x_unit = axes.x_axis.get_unit_size()
        y_unit = axes.y_axis.get_unit_size()

        width = width_ratio * x_unit * x_step

        # Create a list of rectangles arranged side by side,
        # one for each x value
        rects = []
        epsilon = 1e-8
        for x, y in zip(xs, values):
            rect = Rectangle(
                width=width,
                height=max(y * y_unit, epsilon),
                fill_color=fill_color,
                fill_opacity=fill_opacity,
                stroke_color=stroke_color,
                stroke_width=stroke_width,
            )
            rect.move_to(axes.c2p(x + offset * x_step, 0), DOWN)
            rects.append(rect)
        super().__init__(*rects)
        self.axes = axes
        self.xs = xs
        self.set_values(values)

    def set_values(self, values: Iterable[float | int]):
        y_unit = self.axes.y_axis.get_unit_size()
        for rect, value in zip(self, values):
            new_height = value * y_unit
            height = rect.get_top()[1] - rect.get_bottom()[1]
            rescale = new_height / height
            rect.stretch(
                rescale,
                dim=1,
                about_edge=DOWN,
            )



import abc
from typing import Any, Sequence, Type

from manim import *

WOWCHEMY_FOLDER = "docs/static/uploads/media"
QUALITY = "medium_quality"
QUALITY2RES = {
    "high_quality": "1080p60",
    "medium_quality": "720p30",
    "low_quality": "480p15",
}


class ShowcaseMeta(abc.ABCMeta):
    showcase_scenes: Sequence[Type[Scene]] = []

    def __new__(cls, clsname, bases, attrs):
        newclass = super().__new__(cls, clsname, bases, attrs)
        if clsname != "ShowcaseScene":
            cls.showcase_scenes.append(newclass)
        return newclass


class ShowcaseScene(Scene, metaclass=ShowcaseMeta):
    @staticmethod
    @abc.abstractmethod
    def showcasing(self) -> Type[Any]:
        pass



from manim import *

from powermanim.layouts.arrangedbullets import Bullet
from powermanim.showcase.showcasescene import ShowcaseScene
from powermanim.templates.bulletlist import BulletList


class BulletListShowcase(ShowcaseScene):
    def showcasing():
        return BulletList

    def construct(self):
        rows = [
            Bullet("First row", group=0),
            Bullet("Second row", group=1),
            Bullet("Elements:", group=2),
            Bullet("First element", level=1, symbol="(1)", group=3),
            Bullet("Second element", level=1, symbol="(2)", group=4),
            Bullet("Third element", level=1, symbol="(3)", group=5),
        ]
        VGroup(*rows).set_opacity(0.5).scale(0.8)

        bullets = BulletList(
            *rows,
            scale_active=1.2,
        )
        self.add(bullets)

        self.play(bullets.also_next())
        self.play(bullets.also_next())
        self.play(bullets.also_next())
        self.play(bullets.also_next())
        self.play(bullets.also_next())
        self.play(bullets.also_next())
        self.play(bullets.clear())
        self.play(bullets.only_next())
        self.play(bullets.only_next())
        self.play(bullets.only_next())
        self.play(bullets.only_next())
        self.play(bullets.only_next())
        self.play(bullets.only_next())
        self.play(bullets.clear())
        self.wait()



from manim import *


class NumberSlider(VGroup):
    def __init__(
        self,
        x_range=[-2, 2, 0.5],
        length=2,
    ):
        """A class to represent a number together with an associated slider.

        Args:
            x_range (list): A list of 3 floats representing x_min, x_max and x_step.
            length (float): Length of the line.
        """
        self.tracker = ValueTracker(0)

        decimal_number = DecimalNumber(self.tracker.get_value(), num_decimal_places=1)

        number_line = NumberLine(
            x_range=x_range,
            length=length,
            rotation=90 * DEGREES,
        )

        arrow = Arrow(RIGHT / 2, LEFT / 2, buff=0).next_to(number_line.n2p(self.tracker.get_value()), RIGHT)

        arrow.add_updater(lambda obj: obj.become(obj.copy().next_to(number_line.n2p(self.tracker.get_value()), RIGHT)))
        decimal_number.add_updater(
            lambda obj: obj.set_value(
                self.tracker.get_value(),
            ).next_to(number_line, LEFT, buff=MED_LARGE_BUFF)
        )

        super().__init__(decimal_number, number_line, arrow)



import importlib
import sys
from pathlib import Path

# Import all showcased scenes, in order for them to be registered in ShowcaseMeta

all_python_files = [
    f for f in Path(__file__).parent.rglob("*.py") if f.is_file() and not f.name.endswith("__init__.py")
]
for m in all_python_files:
    # TODO: fix this hack to get the module name
    module_name = ".".join(str(m.with_suffix("").relative_to(Path(__file__).parent.parent)).split("/"))

    spec = importlib.util.spec_from_file_location(module_name, m)
    foo = importlib.util.module_from_spec(spec)
    sys.modules["module.name"] = foo
    spec.loader.exec_module(foo)



import typing as T

from manim import *

from powermanim.components.vgrouphighlight import VGroupHighlight
from powermanim.layouts.arrangedbullets import ArrangedBullets


class BulletList(VGroup):
    def __init__(
        self,
        *rows: T.Union[Tex, Text],
        line_spacing: float = MED_LARGE_BUFF * 1.5,
        indent_buff: float = MED_LARGE_BUFF * 1.5,
        left_buff: float = MED_LARGE_BUFF * 2,
        global_shift: float = 0.0,
        inactive_opacity: float = 0.5,
        active_opacity: float = 1.0,
        scale_active: float = 1.0,
    ):
        """A class to represent an highlatable list of items.

        Args:
            rows: A list of items to be displayed.
            line_spacing: The spacing between the rows.
            indent_buff: The spacing between the bullet and the text.
            left_buff: The spacing between the left edge of the rows and the left edge of the screen.
            global_shift: The global_shift to apply to the rows.
            inactive_opacity: The opacity of the inactive items.
            active_opacity: The opacity of the active items.
            scale_active: The scale of the active items.
        """
        self.arranged_list = ArrangedBullets(
            *rows,
            line_spacing=line_spacing,
            indent_buff=indent_buff,
            left_buff=left_buff,
            global_shift=global_shift,
        ).set_opacity(inactive_opacity)

        self.rows = VGroupHighlight(
            *self.arranged_list,
            active_opacity=active_opacity,
            scale_active=scale_active,
            scale_about_edge=LEFT,
        )
        super().__init__(self.rows)

        self.highlighted = 0

    @property
    def ngroups(self) -> int:
        return self.arranged_list.ngroups

    def also_next(self) -> Animation:
        """Highlights also the next item in the list."""
        self.highlighted += 1

        if self.highlighted > self.arranged_list.ngroups:
            raise StopIteration("No more elements to highlight.")

        return self.rows.highlight(indices=list(range(self.highlighted)))

    def only_next(self) -> Animation:
        """Highlights only the next item in the list."""
        if self.highlighted > self.arranged_list.ngroups:
            raise StopIteration("No more elements to highlight.")
        anims = self.rows.highlight(indices=self.highlighted)
        self.highlighted += 1
        return anims

    def clear(self) -> Animation:
        """Clears the list hightlighting."""
        anims = self.rows.highlight(indices=[])
        self.highlighted = 0
        return anims

    def all(self) -> Animation:
        """Highlights all the list."""
        return self.rows.highlight(indices=list(range(len(self.rows))))



from manim import *

from powermanim.showcase.showcasescene import ShowcaseScene
from powermanim.templates.reference import Reference


class ReferenceShowcase(ShowcaseScene):
    def showcasing():
        return Reference

    def construct(self):
        text = (
            r"Moschella et al. “Relative representations enable zero-shot latent space communication”, ICLR 2023 (oral)"
        )
        tex = Tex(text).scale(0.5)
        reference = Reference(text=text)

        self.add(tex)

        self.play(
            Transform(tex, reference[0]),
            rate_func=there_and_back_with_pause,
            run_time=3,
        )



import typing as T
from collections import defaultdict

from manim import *


class MathBullet(MathTex):
    def __init__(
        self,
        *text: str,
        level: int = 0,
        group: T.Optional[int] = None,
        adjustment: float = 0.0,
        symbol: T.Optional[str] = r"\bullet",
        autoplace: bool = True,
        **kwargs,
    ):
        """A class to represent a bullet point.

        Args:
            text: The text to be displayed.
            level: The indent level of the bullet point.
            group: The group the bullet point belongs to, controls the animations.
            adjustment: The adjustment of the bullet.
            symbol: The symbol to be displayed as the bullet.
            autoplace: Whether to automatically place the bullet point.
        """
        self.level = level
        self.adjustment = adjustment
        self.group = group
        self.symbol = symbol
        self.autoplace = autoplace

        first_text, *rest_text = text
        if symbol is not None:
            first_text = rf"{symbol}~{first_text}"

        super().__init__(first_text, *rest_text, **kwargs)

    def indent(self, indent_buff: float = MED_LARGE_BUFF * 1.5):
        """Indent the bullet point.

        Args:
            indent_buff: The spacing between the bullet and the text.
        """
        self.shift(RIGHT * indent_buff * self.level)

    def unindent(self, indent_buff: float = MED_LARGE_BUFF * 1.5):
        """Unindent the bullet point.

        Args:
            indent_buff: The spacing between the bullet and the text.
        """
        self.shift(LEFT * indent_buff * self.intend_level)

    def adjust(self, adjustment: T.Optional[float] = None):
        """Adjust the bullet point.

        Args:
            adjustment: The shift of the bullet.
        """
        if adjustment is None:
            adjustment = self.adjustment
        self.shift(self.adjustment)


class Bullet(MathBullet):
    def __init__(
        self,
        *text: str,
        level: int = 0,
        group: T.Optional[int] = None,
        adjustment: float = 0.0,
        symbol: T.Optional[str] = r"$\bullet$",
        force_inline: bool = False,
        arg_separator="",
        tex_environment="center",
        **kwargs,
    ):
        first_text, *rest_text = text
        if force_inline:
            first_text = r"\mbox{" + first_text
            rest_text = *rest_text, r"}"

        super().__init__(
            first_text,
            *rest_text,
            level=level,
            group=group,
            adjustment=adjustment,
            symbol=symbol,
            arg_separator=arg_separator,
            tex_environment=tex_environment,
            **kwargs,
        )


class ArrangedBullets(VGroup):
    def arrage_rows(self, rows: T.Iterable[T.Union[MathBullet, Bullet, MathTex, Tex, Text]]):
        bullet_rows: T.Iterable[MathBullet] = (
            VGroup(*rows)
            .arrange(DOWN, aligned_edge=LEFT, buff=self.line_spacing)
            .to_edge(LEFT, buff=self.left_buff)
            .shift(self.global_shift)
        )

        for row in bullet_rows:
            row.indent(indent_buff=self.indent_buff)
            row.adjust()

    def __init__(
        self,
        *rows: T.Union[MathBullet, Bullet, MathTex, Tex, Text],
        line_spacing: float = MED_LARGE_BUFF * 1.5,
        indent_buff: float = MED_LARGE_BUFF * 1.5,
        left_buff: float = MED_LARGE_BUFF * 1.5,
        global_shift: float = 0.0,
    ):
        """A VGroup that arranges the rows in a list of bullet points.

        Args:
            rows: A list of items to be displayed.
            line_spacing: The spacing between the rows.
            indent_buff: The spacing between the bullet and the text.
            left_buff: The spacing between the left edge of the rows and the left edge of the screen.
            global_shift: The global_shift of the rows.
        """
        self.line_spacing = line_spacing
        self.indent_buff = indent_buff
        self.left_buff = left_buff
        self.global_shift = global_shift

        rows = [(row if isinstance(row, MathBullet) else Bullet(row)) for row in rows]

        self.arrage_rows((row for row in rows if row.autoplace))

        groups = [row.group for row in rows]

        # If there is a None and aso something else
        if (None in groups) and len(set(groups)) != 1:
            raise ValueError("The groups must be specified for all or no bullets at all.")

        group2bullet = defaultdict(list)
        for i, row in enumerate(rows):
            group = row.group
            if group is None:
                group = i
            group2bullet[group].append(row)

        group_rows = []
        for _, bullets in group2bullet.items():
            group_rows.append(VGroup(*bullets))

        super().__init__(*group_rows)
        self.ngroups = len(self)
